# -*- coding: utf-8 -*-
"""
Created on Wed Feb 28 13:27:12 2024

@author: kony
"""

import numpy as np
from scipy.optimize import linprog


# vytvori polytop z vyplatmi matice symetricke hry
def generate_polytope(A):
    # Předpokládáme, že A je čtvercová matice
    n = A.shape[0]

    # Vytvoření nerovností pro polytop
    # A*z <= 1 a z >= 0
    inequality_lhs = np.vstack([A, -np.eye(n)])
    inequality_rhs = np.hstack([np.ones(n), np.zeros(n)])

    return inequality_lhs, inequality_rhs


# pomocna procedura zjistujici, jestli je vrchol v seznamu
def is_in_visited(vertex, visited_list):
    for visited_vertex in visited_list:
        if np.allclose(vertex, visited_vertex):
            return True
    return False


# pro vrchol najde dosud nenavstiveneho souseda
def find_neighboring_vertices(current_vertex, inequality_lhs, inequality_rhs, visited):
    n = inequality_lhs.shape[1]

    # Identifikace aktuálních omezení
    current_constraints = [i for i in range(inequality_lhs.shape[0]) 
                            if np.isclose(inequality_lhs[i] @ current_vertex, inequality_rhs[i])]

    constraints_to_loosen = []
    constraints_to_add = []
    
    #jestli neco neni reprezentovano 2x
    for i in current_constraints:
        
        if (i + n) in current_constraints:
            constraints_to_loosen += [i,i+n]
        
        if i == n:
            break
        

    #jestli neco neni reprezentovano vubec
    # for i in range(n):        
    #     if  not( (i + n) in current_constraints) and  not( (i + n) in current_constraints):
    #         constraints_to_add += [i,i+n]

    

    if not constraints_to_loosen:
        # nic neni representovano 2x
        # bud jsme v 0 nebo v NE
        if sum(current_vertex) == 0:
            constraints_to_loosen = current_constraints
        else:
            return current_vertex, True

    # Procházení ostatních omezení, aby se našli sousední vrcholy
    for i in range(inequality_lhs.shape[0]):
        if i not in current_constraints:
            for j in constraints_to_loosen:
                # Vytvoření nové sady omezení
                new_constraints = current_constraints[:]
                new_constraints.remove(j)
                new_constraints.append(i)

                # Vytvoření lineárního programu
                A_eq = inequality_lhs[new_constraints, :]
                b_eq = inequality_rhs[new_constraints]

                # Řešení lineárního programu
                res = linprog(np.zeros(n), A_ub=inequality_lhs, b_ub=inequality_rhs, A_eq=A_eq, b_eq=b_eq, method='highs')

                # Kontrola, zda je výsledek novým vrcholem
                # if res.success and np.all(A_eq @ res.x <= b_eq) and not np.array_equal(res.x, current_vertex):
                if res.success and not is_in_visited(res.x, visited):
                    return res.x, False


print("Testovací matice A")
# Testovací matice A
A = np.array([[0, 3, 0],
              [0, 0, 3],
              [2, 2, 2]])

inequality_lhs, inequality_rhs = generate_polytope(A)

vertex = np.array([0,0,0])
visited = []
finish = False
while not finish:
    visited += [vertex]
    print(vertex)
    vertex, finish = find_neighboring_vertices(vertex, inequality_lhs, inequality_rhs,visited)



print("Arena bez 1")
# Testovací matice A (Arena bez 1)
A = np.array([[0, 0, 0, 1, 3, 3],
              [0, 0, 0, 4, 1, 4],
              [0, 0, 0, 5, 5, 1],
              [5, 2, 1, 0, 0, 0],
              [3, 5, 1, 0, 0, 0],
              [3, 2, 5, 0, 0, 0]])

inequality_lhs, inequality_rhs = generate_polytope(A)

vertex = np.array([0,0,0,0,0,0])
visited = []
finish = False
while not finish:
    visited += [vertex]
    print(vertex)
    vertex, finish = find_neighboring_vertices(vertex, inequality_lhs, inequality_rhs,visited)
    

    
print("Arena bez 2")
# Testovací matice A (Arena bez 2)
A = np.array([[0, 0, 0, 1, 2, 2],
              [0, 0, 0, 4, 1, 4],
              [0, 0, 0, 5, 5, 1],
              [5, 2, 1, 0, 0, 0],
              [4, 5, 1, 0, 0, 0],
              [4, 2, 5, 0, 0, 0]])

inequality_lhs, inequality_rhs = generate_polytope(A)

vertex = np.array([0,0,0,0,0,0])
visited = []
finish = False
while not finish:
    visited += [vertex]
    print(vertex)
    vertex, finish = find_neighboring_vertices(vertex, inequality_lhs, inequality_rhs,visited)


print("Arena cela")
# Testovací matice A (Arena cela)
A = np.array([[0, 0, 0, 0, 1, 2, 2, 2],
              [0, 0, 0, 0, 3 ,1, 3, 3],
              [0, 0, 0, 0, 4, 4, 1, 4],
              [0, 0, 0, 0, 5, 5, 5, 1],
              [5, 3, 2, 1, 0, 0, 0, 0],
              [4, 5, 2, 1, 0, 0, 0, 0],
              [4, 3, 5, 1, 0, 0, 0, 0],
              [4, 3, 2, 5, 0, 0, 0, 0]])
inequality_lhs, inequality_rhs = generate_polytope(A)

vertex = np.array([0,0,0,0,0,0,0,0])
visited = []
finish = False
while not finish:
    visited += [vertex]
    print(vertex)
    vertex, finish = find_neighboring_vertices(vertex, inequality_lhs, inequality_rhs,visited)


